# frozen_string_literal: true

User.class_eval do
  field :main_applicant_id, type: BSON::ObjectId
  field :reservation_holder_id, type: BSON::ObjectId
  field :lessor_id, type: BSON::ObjectId
  field :gender, type: String
  field :id_cristal, type: String

  index({main_applicant_id: 1}, {background: true})
  index({reservation_holder_id: 1}, {background: true})
  index({lessor_id: 1}, {background: true})
  index({gender: 1}, {background: true})
  index({id_cristal: 1}, {background: true})

  before_save :hermes_before_actions
  after_save :hermes_after_actions
  after_create :hermes_after_create_actions

  has_many(
    :notifications,
    class_name: '::NotificationCenter::Notification',
    foreign_key: :targetable_id,
    dependent: :destroy,
    order: 'created_at DESC'
  )

  ###############################
  #### Notification #############
  include ::Concerns::NotificationCenter::Notifiable

  def self.targets_selector(hermes_user, opt = {})
    [hermes_user].compact
  end

  def notification_mailer
    ::Hermes::NotificationCenter::HermesMailer
  end

  def notifiable_title(key, opt={})
    I18n.t("#{key}.title")
  end

  def notifiable_body(key, opt={})
    I18n.t("#{key}.body")
  end

  def notifiable_path
    ""
  end

  ###############################
  ###############################

  def has_role?(role)
    ((roles || []) + (auth_roles || [])).include? role
  end

  # user.has_roles?('role1', 'role2') or user.has_roles?(%[role1 role2])
  def has_roles?(*args)
    if (args.count == 1) && args.first.is_a?(Array)
      return (((roles || []) + (auth_roles || [])) & args.first).any?
    end
    (((roles || []) + (auth_roles || [])) & args).any?
  end

  def main_applicant
    Hermes::Applicant.find_by(id: main_applicant_id)
  end

  def reservation_holder
    Hermes::ReservationHolder.find(reservation_holder_id)
  end

  def lessor
    Hermes::Lessor.find(lessor_id)
  rescue
    nil
  end

  def lessor=(les)
    self.lessor_id = les.id
  end

  def create_user_on_identity_provider
  end

  def create_new_user_on_identity_provider(already_check = 0 )
    if provider && 
      create_user_data && 
      !create_user_data.empty? && 
      identity_provider = IdentityProvider.find_by(key: provider)
      
      @create_user_response = send_create_user_request_to_identity_provider(identity_provider, create_user_data)
      update_user_roles_beiam(already_check)
      
      return false unless [200, 201, 409].include?(@create_user_response.code)
    end
  end
  
  def update_user_roles_beiam(already_check = 0)
    return unless identity_provider = IdentityProvider.find_by(key: provider)
    url_get = URI.escape(identity_provider.reload.add_roles_user_url.sub!('{login}', self.login).sub!('{role}', ""))
    token = get_beiam_token(identity_provider)
    data = {headers: {'Authorization': "Bearer #{token}"}}
    response = HTTParty.get(url_get, data) 

    if already_check >= 2
      Log.perform_later(:code => "BEIAM-LOGIN-ERROR", :action => "Impossible de mettre à jour les roles de #{email} sur beiam", user_id: user.id)
      raise "Update user auth roles on beiam failed for login #{login}"
      return
    end 

    if response.code == 404
      create_new_user_on_identity_provider(already_check + 1)
      return
    end

    roles = response.code == 200 ? JSON.parse(response.body).map { |r| r['name']} : []
    
    added_roles = (self.auth_roles || []) - (roles || [])
    deleted_roles = (roles || []) - (self.auth_roles  || []) 
    
    added_roles.each do |role|
      send_add_roles_beiam(identity_provider, role, true)
    end
    
    deleted_roles.each do |role|
      send_add_roles_beiam(identity_provider, role, false)
    end
  end
  
  def send_create_user_request_to_identity_provider(identity_provider, create_user_data = {}, update = false)
    url = identity_provider.create_user_url
    url += "/" + login if update
    
    data = {
      headers: identity_provider.create_user_headers,
      body: create_user_data.to_json,
    }
  
    update ? HTTParty.put(url, data) : HTTParty.post(url, data) 
  end
  
  def send_add_roles_beiam(identity_provider, role, add)
    token = get_beiam_token(identity_provider)
    url = URI.escape(identity_provider.reload.add_roles_user_url.sub!('{login}', self.login).sub!('{role}', role))
    data = {headers: {'Authorization': "Bearer #{token}"}}
    
    add ? HTTParty.put(url, data) : HTTParty.delete(url, data)
  end

  def get_beiam_token(identity_provider)
    RestClient.post(
      identity_provider.account_service_url,
      headers: { "Content-Type" => "application/x-www-form-urlencoded" },
      client_id: identity_provider.account_service_client_id,
      client_secret: identity_provider.account_service_client_secret,
      grant_type: 'client_credentials'
    ).json_to_h[:access_token]
  end

  def has_lessor?
    lessor_id.present?
  end

  def update_rights
    deleted_group_ids = (group_ids_was || []) - (group_ids || []) 
    added_group_ids = (group_ids || []) - (group_ids_was || []) 
    
    deleted_group_ids.each do |id|
      ur = Hermes::UserRight.find_by(user_id: self.id, group_id: id.to_s)
      ur.destroy if ur
    end
    
    added_group_ids.each do |id|
      if data = create_right_data(id)
        ur = Hermes::UserRight.find_by(user_id: self.id, group_id: id.to_s)
        Hermes::UserRight.create(data) unless ur
      end
    end
  end

  def create_right_data(id)
    return unless group = Group.find(id)

    right = case group.name
      when /^entity_manager_/
        'administrator'
      when /^entity_supervisor_/
        'user'
      when /^entity_/
        'reader'
      else
        nil
      end

      if right 
        matching_type = {
          administrator: 'manager_group_id',
          user: 'supervisor_group_id',
          reader: 'group_id',
        }

        query = {}
        query[matching_type[right.to_sym]] = group.id
        lessor = Hermes::Lessor.find_by(query) 
        entity = Hermes::Entity.find_by(query) if !lessor
        
        o = entity || lessor
        return unless o
        return {
          user_id: self.id,
          entity_id: lessor ? nil : o.id ,
          lessor_id: lessor ? o.id  : nil,
          group_id: group&.id.to_s,
          title: o.title,
          right: right,
          type: lessor ? 'lessor' : 'entity'
        }
      end
    nil
  end
  
  def hermes_before_actions
    check_roles if auth_roles_changed?
  end

  def hermes_after_actions
    update_rights if group_ids_changed?
    update_user_roles_beiam if auth_roles_changed?
  end

  def check_roles
    self.auth_roles = self.auth_roles - ['entreprise', 'bailleur'] unless (['cre', 'rdsl'] & self.auth_roles).blank?
  end 

  def hermes_after_create_actions
    notify(key: NOTIFICATION_KEYS[:user][:create], channels: [])
  end
end
