# frozen_string_literal: true

require 'rails_helper'

# %w[
#   applicant
#   housing_request
# ].each { |factory_name|require Hermes::Engine.root.join('spec', 'factories', "#{factory_name.pluralize}_factory").to_s }

RSpec.describe Hermes::HousingRequest, type: :model do
  describe 'basic controls' do
    let!(:user1) {create(:user)}
    let!(:applicant1) do
      create(
        :applicant,
        user: user1,
        monthly_charges: {}
      )
    end
    
    it 'has a valid factory' do
      expect(FactoryBot.build(:housing_request, main_applicant: applicant1)).to be_valid
    end
  end
  
  context 'with simple setup' do
    before :each do
      $redis.flushall
    end
    let!(:user1) {create(:user)}
    let!(:user2) {create(:user)}
    let!(:user3) {create(:user)}
    let!(:applicant1) {
      create(
        :applicant,
        user: user1,
        monthly_charges: {},
      )
    }
    let!(:applicant2) {
      create(
        :applicant,
        user: user2
      )
    }
    let!(:applicant3) {
      create(
        :applicant,
        user: user3
      )
    }
    
    let!(:housing_request1) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: applicant1
      )
    end
    
    let!(:housing_request2) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: applicant1,
        coapplicants: [
          {rel_type: 'Conjoint (e)', id: applicant2.id.to_s},
          {rel_type: 'Parent (e)', id: applicant3.id.to_s}
        ]
      )
    end
    
    describe 'Eligibility query are valid' do
      it 'should include right typology ranges' do
        typology_range1 = housing_request1.calculate_typology_range
        typology_range2 = housing_request2.calculate_typology_range
        range1_to_i = typology_range1.map {|typology| Integer(typology.sub('T', ''))}
        range2_to_i = typology_range2.map {|typology| Integer(typology.sub('T', ''))}
        
        (1..2).each do |n|
          (1..2).each do |i|
            expect(binding.local_variable_get("range#{n}_to_i".to_sym)).to include i
          end
        end
        expect(range1_to_i.max).to be 2
        expect(range2_to_i).to include 3
        expect(range2_to_i.max).to be 3
      end
      
      it 'eligibility should incorpore typology' do
        typology_filter1 = housing_request1.typology_filter
        expect(typology_filter1[:typology][:$in]).to include 'T1'
        expect(typology_filter1[:typology][:$in]).to include 'T2'
      end
      
      describe 'edit action dont duplicate coapplicants' do
        before :each do
          housing_request2.coapplicants = [
            {rel_type: 'Conjoint (e)', id: applicant3.id}
          ]
          housing_request2.seeking_elevator = false
          housing_request2.save
          housing_request2.reload
          housing_request2.coapplicants = [{rel_type: 'M', id: applicant3.id.to_s}]
          housing_request2.save
          applicant3.destroy
          housing_request2.reload
          housing_request2.forcing_duplicate = true
          housing_request2.save
          housing_request2.reload
        end
        it 'coapplicants doesnt change' do
          expect(housing_request2.reload.applicant_ids.length).to eq(0)
        end
      end
      
      describe 'delete applicantId delete coapplicants' do
        before :each do
          housing_request2.applicant_ids = []
          housing_request2.save
        end
        it 'coapplicants doesnt change' do
          expect(housing_request2.reload.coapplicants.length).to eq(0)
        end
      end
    end
    
    describe "#update_seeked_location" do
      before :each do
      $redis.flushall
      end
      let(:list) do
        {'nom': 'Fréjus', 'code': '83061', 'codeDepartement': '81', 'codesPostaux': ['83900', '83370', '']}
      end
      let(:lists) do
        [
          {'nom': 'Paris', 'code': '75001', 'codeDepartement': '75', 'codesPostaux': ['75006', nil]},
          {'nom': 'Meaux', 'code': '77000', 'codeDepartement': '77'},
          {'nom': 'Paris', 'code': '75002', 'codeDepartement': '75', 'codesPostaux': ['75006', '93000']}
        ]
      end
      let(:broken_list) do
        {'code': '77013', 'codePostal': '77013', 'foo': 'bir'}
      end
      
      it 'sets seeked fields on save' do
        housing_request1.seeked_localisation_list = [list]
        expect {housing_request1.save}.to change {housing_request1.seeked_departments}.from([]).to(['81'])
                                            .and change {housing_request1.truly_seeked_departments}.from([]).to(['81', '83'])
                                                   .and change {housing_request1.seeked_district_postal_codes}.from([]).to(['83370', '83900'])
                                                          .and change {housing_request1.seeked_districts}.from([]).to(['Fréjus'])
      end
      
      it 'ensures the sorted uniqueness of seeked fields' do
        housing_request1.seeked_localisation_list = lists + [list]
        housing_request1.save
        expect(housing_request1.seeked_departments).to eq ['75', '77', '81']
        expect(housing_request1.truly_seeked_departments).to eq ['75', '77', '81', '83', '93']
        expect(housing_request1.seeked_district_postal_codes).to eq ['75006', '83370', '83900', '93000']
        expect(housing_request1.seeked_districts).to eq ['Fréjus', 'Meaux', 'Paris']
      end
      
      it 'returns error' do
        housing_request1.seeked_localisation_list = [broken_list]
        housing_request1.save
        expect(housing_request1.errors.full_messages.first).to match /non-conform seeked_localisation_list/
      end
    end
    
    describe 'Relationships are valid' do
      before :each do
        $redis.flushall
      end
      it 'should has correct housing_requests situation' do
        expect(applicant1.first_active_request.id).to eq(housing_request2.id)
        expect(applicant1.housing_requests.length).to eq(2)
        expect(applicant1.housing_requests.actives.length).to eq(1)
      end
      
      it 'should has correct values' do
        revenues_sum = (applicant1.monthly_revenues || 0) + (applicant2.monthly_revenues || 0)
        expect(housing_request2.monthly_revenues).to eq(revenues_sum)
      end
      
      it 'should has correct coapplicants' do
        coapplicants = housing_request2.coapplicants
        expect(coapplicants.length).to eq(2)
        expect(coapplicants.map {|coapplicant| coapplicant[:rel_type]}).to include('Conjoint (e)').and include('Parent (e)')
        expect(coapplicants.map {|coapplicant| coapplicant[:id]}).to include(applicant2.id).and include(applicant3.id)
      end
      
      describe 'after coapplicants update' do
        before :each do
          housing_request2.coapplicants = [
            {rel_type: 'Conjoint (e)', id: applicant3.id}
          ]
          housing_request2.save
        end
        
        it 'should has updated with valid data' do
          r_request = housing_request2.reload
          expect(r_request.applicant_ids.length).to be 1
          expect(r_request.applicant_ids.first).to eq applicant3.id
          expect(r_request.coapplicants.first[:rel_type]).to eq 'Conjoint (e)'
          expect(Hermes::ApplicantLink.where(housing_request_id: r_request.id).count).to be 1
        end
      end
    end
    
    describe 'applicant should have only public offer as eligibility when last active request becomes inactive' do
      it 'should has first active request with valid data' do
        expect(housing_request2.main_applicant.housing_requests.actives.count).to eq 1
        expect(housing_request2.is_active).to be true
        expect(housing_request2.main_applicant.first_active_request.id.to_s).to eq housing_request2.id.to_s
      end
      
      context 'is_active is set to false' do
        before :each do
          housing_request2.update(is_active: false)
        end
        
        it 'should has is_active == false' do
          expect(housing_request2.reload.is_active).to be false
        end
        
        it 'should has its main_applicant.eligibility == {is_public: true}' do
          expect(housing_request2.reload.main_applicant.eligibility_query).to eq '{"is_public":true}'
        end
      end
    end
    
    describe "#generate_als_id" do
      before :each do
        $redis.flushall
      end
      let(:housing_request3) do
        build(
          :housing_request,
          is_active: true,
          main_applicant: applicant1
        )
      end
      
      it {expect {housing_request3.save}.to change {housing_request3.als_id}.from(nil).to '0000000003'}
    end
  end
  
  context 'with large quantity' do
    nb = 20
    nb.times do |i|
      let!("applicant#{i}") {create(:applicant)}
      let!("request#{i}") {create(:housing_request, main_applicant: send("applicant#{i}"))}
    end
    
    describe 'als_id is set' do
      let(:all_requests) {Hermes::HousingRequest.only(:als_id).where(als_id: {:$ne => nil})}
      
      it "has #{nb} entities" do
        expect(all_requests.count).to be nb
      end
      
      let(:als_ids) {all_requests.pluck(:als_id).uniq.compact}
      
      it 'produces valid length' do
        expect(als_ids.map(&:length)).to all(be(10))
      end
      
      it 'has no duplicates' do
        expect(
          als_ids.map {|als_id| all_requests.where(als_id: als_id).count}
        ).to all(be(1))
      end
      
   
    end
  end
  
  describe "update_from_sne" do
    let!(:housing_request) { create(:sne_integrate_request) }

    it 'should add coapplicants to request' do
      # The sne request containt a coapplicant, it should populate to housing_request
      expect(housing_request.coapplicants).to be_blank
      expect(housing_request.update_request_from_sne).to be(true)
      expect(housing_request.reload.coapplicants).not_to be_blank
    end

    it 'should update existing coapplicants' do
      # The coapplicant from sne is Tim PLATO 1989-01-01 (only first_name is different)
      coapplicant = FactoryBot.create(:applicant, first_name: 'Arthur', last_name: 'PLATO', birth_date: '1989-01-01')
      housing_request.coapplicants = [{ rel_type: 'L', id: coapplicant.id.to_s }]
      housing_request.save
      housing_request.reload
      expect(housing_request.coapplicants).not_to be_blank
      expect(housing_request.update_request_from_sne).to be(true)
      expect(housing_request.reload.coapplicants).not_to be_blank

      found_coapplicant = Hermes::Applicant.find housing_request.coapplicants.first[:id]
      expect(found_coapplicant).to eq(coapplicant)
      expect(found_coapplicant.first_name).to eq('Tim')
    end

    it 'should delete coapplicants if not exist on sne' do
      # Only Tim PLATO exist on sne
      coapplicant1 = FactoryBot.create(:applicant, first_name: 'Tim', last_name: 'PLATO', birth_date: '1989-01-01')
      coapplicant2 = FactoryBot.create(:applicant, first_name: 'Arthur', last_name: 'DUPONT', birth_date: '1989-01-01')
      housing_request.coapplicants = [{ rel_type: 'L', id: coapplicant1.id.to_s }, { rel_type: 'L', id: coapplicant2.id.to_s }]
      housing_request.save
      housing_request.reload
      expect(housing_request.coapplicants.count).to eq(2)
      expect(housing_request.update_request_from_sne).to be(true)
      expect(housing_request.coapplicants.count).to eq(1)

      found_coapplicant = Hermes::Applicant.find housing_request.coapplicants.first[:id]
      expect(found_coapplicant).to eq(coapplicant1)
    end

    it 'should delete and create coapplicants if more than 1 fields are differents' do
      # The coapplicant from sne have different first_name and last_name
      coapplicant = FactoryBot.create(:applicant, first_name: 'Arthur', last_name: 'DUPONT', birth_date: '1989-01-01')
      housing_request.coapplicants = [{ rel_type: 'L', id: coapplicant.id.to_s }]
      housing_request.save
      housing_request.reload

      expect(housing_request.update_request_from_sne).to be(true)
      housing_request.reload
      expect(housing_request.coapplicants.count).to eq(1)

      found_coapplicant = Hermes::Applicant.find housing_request.coapplicants.first[:id]
      expect(found_coapplicant).not_to eq(coapplicant)
    end
  end

  describe "After actions" do
    let!(:user) { create(:user) }
    let!(:lessor) { create(:lessor) }
    let!(:entity) { create(:entity, lessor: lessor) }
    let!(:housing_offer) { create(:housing_offer, lessor: lessor) }
    let!(:applicant) { create(:applicant, user: user, entity: entity) }  
    let!(:housing_request) { create(:housing_request, is_active: true, main_applicant: applicant, request_status: 'COMPLETE') }
    let!(:housing_application) { create(:housing_application, housing_offer: housing_offer, housing_request: housing_request, status: 'TRANSMIS BAILLEUR' ) }
    
    it "should create public shares if request is applicated" do
      hrs = Hermes::HousingRequestShare.find_by(housing_request_id: housing_request.id, lessor_id: lessor.id)
      expect(hrs).not_to be_nil
    end

    it "should create and link housing_request_folder" do
      folders = Hermes::HousingRequestFolder.where(housing_request_id: housing_request.id)
      expect(housing_request.reload.housing_request_folder_id).not_to be_nil
      expect(folders.count).to eq(1)
    end
  end

  describe "Entity validations" do
    let!(:user) { create(:user) }
    let!(:lessor) { create(:lessor) }
    let!(:entity) { create(:entity, lessor: lessor) }
    let!(:lessor2) { create(:lessor) }
    let!(:entity2) { create(:entity, lessor: lessor2, auto_request_validation: true) }
    let!(:applicant) { create(:applicant, user: user, entity: entity) }
    let!(:housing_request) { create(:housing_request, is_active: true, main_applicant_id: applicant.id, request_status: 'COMPLETE') }
  
    it "should automaticly validate request" do
      applicant.update(entity: entity2)
    
      expect(housing_request.reload.is_validated?).to eq(true)
    end
  
    it "should not automaticly validate request" do
      entity2.update(auto_request_validation: false)
      applicant.update(entity: entity2)
    
      expect(housing_request.reload.is_validated?).to eq(false)
    end
  
    it 'should auto set entity_status' do
      hr = FactoryBot.create(:housing_request, is_active: true, main_applicant: applicant)
     # expect(hr.entity_status).to eq(nil)

      hr.update(consent_level: 4)
      # expect(hr.entity_status).to eq('EN COURS DE SAISIE')
    end
  end
  describe "Entity_status workflow" do
    let!(:user) { create(:user) }
    let!(:lessor) { create(:lessor) }
    let!(:entity) { create(:entity, lessor: lessor) }
    let!(:lessor2) { create(:lessor) }
    let!(:entity2) { create(:entity, lessor: lessor2, auto_request_validation: false) }
    let!(:applicant) { create(:applicant, user: user, entity: entity2) }
    let!(:housing_request) { create(:housing_request, is_active: true, main_applicant: applicant, consent_level: 4) }
  

    describe 'has initial entity_status' do
      it "should have good initial entity_status" do
      
        expect(housing_request.entity_status).to eq('EN COURS DE SAISIE')
      end
    end
    describe 'has next entity_status' do
      it "should have good initial entity_status" do
        ap housing_request.housing_request_folder
        housing_request.request_status = "ELIGIBLE"
        housing_request.save
        expect(housing_request.reload.entity_status).to eq('A VALIDER')
      end
    end
    describe 'has next entity_status' do
      it "should have good initial entity_status" do
        ap housing_request.housing_request_folder
        entity2.auto_request_validation = true
        entity2.save
        housing_request.reload.request_status = "ELIGIBLE"
        housing_request.save
        expect(housing_request.reload.entity_status).to eq('EN COURS')
    
      end
  
    end
  end

  describe 'Coapplicants Validations' do
    let(:main_applicant) { create(:applicant) }
    let(:housing_request) { create(:housing_request, main_applicant: main_applicant, is_active: true) }
    let(:coapplicant) { create(:applicant) }

    it 'can add coapplicants with blank rel_type' do
      housing_request.update(coapplicants: [{ id: coapplicant.id.to_s, rel_type: '' }])
      expect(housing_request.reload.coapplicants.first[:rel_type]).to be_blank
    end

    it 'can not add coapplicants with blank rel_type if form_status is complete' do
      housing_request.update(form_status: 'complete', coapplicants: [{ id: coapplicant.id.to_s, rel_type: '' }])
      expect(housing_request.errors.any?).to be(true)
      field, error = housing_request.errors.first
      expect(field).to eq(:coapplicants)
      expect(error).to include("rel_type can't be blank for coappllicants")
    end

    it 'can add coapplicants with rel_type if form_status is complete' do
      housing_request.update(form_status: 'complete', coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C' }])
      expect(housing_request.errors.any?).to be(false)
    end
  end

  describe "inherit readers and writers" do
    let!(:user_reader) {create(:user)}
    let!(:user_writer) {create(:user)}
    let!(:user) {create(:user)}
    let!(:entity) {create(:entity)}
    let!(:applicant) do
      create(
        :applicant,
        user: user,
        monthly_charges: {},
        entity_id: entity.id.to_s,
        readers: [user_reader.id.to_s],
        writers: [user_writer.id.to_s]
      )
    end
    let(:housing_request) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: applicant
      )
    end

    it 'from applicant and entity' do
      expect(housing_request.reload.readers.include?(user_reader.id.to_s)).to be true
      expect(housing_request.reload.writers.include?(user_writer.id.to_s)).to be true

      [entity.group_id, entity.supervisor_group_id, entity.manager_group_id].each do |eg|
        expect(housing_request.reload.readers.include?("g_"+eg)).to be true
      end
      applicant.update(entity_id: nil)

      [entity.group_id, entity.supervisor_group_id, entity.manager_group_id].each do |eg|
        expect(housing_request.reload.readers.include?("g_"+eg)).to be false
      end
    end
  end

  describe 'typology_range & min_surface' do
    let(:applicant) { create(:applicant) }
    let(:coapplicant) { create(:applicant) }
    let(:hr) { create(:housing_request, is_active: true, main_applicant: applicant ) }

    describe 'right for' do 
      pre_mapping = {
        '1 personne seule' => { typology: 1..2, min_surface: 9, less_limited_dependent_nb: 0, coapplicant_nb: 0 },
        '2 personnes' => { typology: 1..3, min_surface: 16, less_limited_dependent_nb: 0, coapplicant_nb: 1 },
        '2 personnes + 1 enfant à charge' => { typology: 2..4, min_surface: 25, less_limited_dependent_nb: 1, coapplicant_nb: 1 },
        '1 personne seule + 1 enfant à charge' => { typology: 2..3, min_surface: 16, less_limited_dependent_nb: 1, coapplicant_nb: 0 },
        '2 personnes + 2 enfants à charge' => { typology: 3..5, min_surface: 34, less_limited_dependent_nb: 2, coapplicant_nb: 1 },
        '1 personne seule + 2 enfants à charge' => { typology: 2..4, min_surface: 25, less_limited_dependent_nb: 2, coapplicant_nb: 0 },
        '2 personnes + 3 enfants à charge' => { typology: 4..6, min_surface: 43, less_limited_dependent_nb: 3, coapplicant_nb: 1 },
        '1 personne seule + 3 enfants à charge' => { typology: 3..5, min_surface: 34, less_limited_dependent_nb: 3, coapplicant_nb: 0 },
        '2 personnes + 4 enfants à charge' => { typology: 5..7, min_surface: 52, less_limited_dependent_nb: 4, coapplicant_nb: 1 },
        '1 personne seule + 4 enfants à charge' => { typology: 4..6, min_surface: 43, less_limited_dependent_nb: 4, coapplicant_nb: 0 },
        '2 personnes + 5 enfants à charge' => { typology: 6..8, min_surface: 61, less_limited_dependent_nb: 5, coapplicant_nb: 1 },
        '1 personne seule + 5 enfants à charge' => { typology: 5..7, min_surface: 52, less_limited_dependent_nb: 5, coapplicant_nb: 0 },
        '2 personnes + 6 enfants à charge' => { typology: 6..9, min_surface: 70, less_limited_dependent_nb: 6, coapplicant_nb: 1 },
        '1 personne seule + 6 enfants à charge' => { typology: 6..8, min_surface: 61, less_limited_dependent_nb: 6, coapplicant_nb: 0 },
        '2 personnes + 7 enfants à charge' => { typology: 6..10, min_surface: 79, less_limited_dependent_nb: 7, coapplicant_nb: 1 },
        '1 personne seule + 7 enfants à charge' => { typology: 7..9, min_surface: 70, less_limited_dependent_nb: 7, coapplicant_nb: 0 },
        '2 personnes + 8 enfants à charge' => { typology: 6..11, min_surface: 88, less_limited_dependent_nb: 8, coapplicant_nb: 1},
       }

      mappings = pre_mapping.each do |k, a|
        pre_mapping[k][:typology] = (a[:typology]).to_a.map do |i|
          "T#{i.to_s}" 
        end
      end

      mappings.each do |k, mapping|
        context k do 
          it 'get the good value' do
            hr.dependents = Array.new(mapping[:less_limited_dependent_nb], "dependent") 
            hr.applicant_ids = mapping[:coapplicant_nb] == 1 ? [coapplicant.id] : [] 
            expect(hr.calculate_typology_range).to eq mapping[:typology]
            expect(hr.calculate_min_surface).to eq mapping[:min_surface]
          end
        end
      end
      
      it 'unborn_children activ' do
        birth_date = (DateTime.now + 150.days).to_s
        hr.unborn_children = [
          {"anticipated_birth"=> birth_date, "nb_child_birth"=>"1"},
          {"anticipated_birth"=> birth_date, "nb_child_birth"=>"1"},
        ]
        hr.save
        expect(hr.calculate_unborn_children_nb).to eq 2
        expect(hr.typology_range).to eq mappings['1 personne seule + 2 enfants à charge'][:typology]
        expect(hr.min_surface).to eq mappings['1 personne seule + 2 enfants à charge'][:min_surface]
        expect(hr.family_size).to eq 3
        expect(hr.main_applicant.eligibility_query.json_to_h[:typology]["$in"]).to eq mappings['1 personne seule + 2 enfants à charge'][:typology]
      end

      it 'unborn_children inactiv' do
        birth_date = (DateTime.now + 200.days).to_s
        hr.unborn_children = [
          {"anticipated_birth"=> birth_date, "nb_child_birth"=>"1"},
          {"anticipated_birth"=> birth_date, "nb_child_birth"=>"1"},
        ]
        hr.save
        expect(hr.calculate_unborn_children_nb).to eq 0
        expect(hr.typology_range).to eq mappings['1 personne seule'][:typology]
        expect(hr.min_surface).to eq mappings['1 personne seule'][:min_surface]
        expect(hr.family_size).to eq 1
        
        expect(enqueued_jobs.select {|j| j[:job] == UnbornChildrenJob}.length).to eq 2
        
        Timecop.freeze(DateTime.now + 50.days)
        drain_all_jobs
        expect(hr.calculate_unborn_children_nb).to eq 2
        expect(hr.reload.typology_range).to eq mappings['1 personne seule + 2 enfants à charge'][:typology]
        expect(hr.reload.min_surface).to eq mappings['1 personne seule + 2 enfants à charge'][:min_surface]
        expect(hr.family_size).to eq 3
        Timecop.return
      end

    end

    describe "inherit birth date from applicant" do
      let!(:applicant) do
        create(
          :applicant,
          birth_date: '2000-01-01'
        )
      end
      let(:housing_request) do
        create(
          :housing_request,
          is_active: true,
          main_applicant: applicant
        )
      end
  
      it 'at creation' do
        expect(housing_request.requester_birth_date).to eq(applicant.birth_date)
      end

      it 'changed applicant birth date' do
        applicant.birth_date = "1900-01-01"
        applicant.save
        applicant.reload
        expect(housing_request.requester_birth_date).to eq(applicant.birth_date)
      end
    end

  end

  describe 'calculate category_type for' do 

    context 'one person' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant)}

      it 'as good category' do 
        expect(hr.reload.category_type).to eq(1)
      end 
    end 

    context 'two people (not young couple)' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 30.years)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, applicant_ids: [coapplicant.id.to_s])}
      
      it 'as good category' do 
        expect(hr.reload.category_type).to eq(2)
      end 
    end 
    
    context 'one person with handicap' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, disability_details: [{lol: 'test'}])}
      
      it 'as good category' do 
        expect(hr.reload.handicap?).to eq(true)
        expect(hr.reload.category_type).to eq(2)
      end 
    end 

    context '3 people' do 
      let(:applicant) { create(:applicant)}
      let(:coapplicant) { create(:applicant)}
      let(:coapplicant2) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, applicant_ids: [coapplicant.id.to_s, coapplicant2.id.to_s])}
      
      it 'as good category' do 
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context '1 person with on dependent' do 
      let(:applicant) { create(:applicant, family_situation: 'Célibataire')}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{rel_type: 'E'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context 'young couple' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 20.years)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}])}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context '2 people with handicap' do 
      let(:applicant) { create(:applicant)}
      let(:coapplicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, applicant_ids: [ coapplicant.id.to_s ], disability_details: [{lol: 'test'}])}
      
      it 'as good category' do 
        expect(hr.reload.handicap?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context '4 people' do 
      let(:applicant) { create(:applicant)}
      let(:coapplicant) { create(:applicant)}
      let(:coapplicant2) { create(:applicant)}
      let(:coapplicant3) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, applicant_ids: [coapplicant.id.to_s, coapplicant2.id.to_s, coapplicant3.id.to_s])}
      
      it 'as good category' do 
        expect(hr.reload.category_type).to eq(4)
      end 
    end 

    context '1 person with 2 dependents' do 
      let(:applicant) { create(:applicant, family_situation: 'Célibataire')}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{rel_type: 'E'}, {rel_type: 'enfant à charge'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(true)
        expect(hr.reload.category_type).to eq(4)
      end 
    end 

    context '3 people with handicap' do 
      let(:applicant) { create(:applicant)}
      let(:coapplicant) { create(:applicant)}
      let(:coapplicant2) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, applicant_ids: [ coapplicant.id.to_s, coapplicant2.id.to_s ], disability_details: [{lol: 'test'}])}
      
      it 'as good category' do 
        expect(hr.reload.handicap?).to eq(true)
        expect(hr.reload.category_type).to eq(4)
      end 
    end 

    context '1 person with 2 dependents' do 
      let(:applicant) { create(:applicant, family_situation: 'Célibataire')}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{rel_type: 'E'}, {rel_type: 'enfant à charge'}, {rel_type: 'garde alternée'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(true)
        expect(hr.reload.category_type).to eq(5)
      end 
    end 

    ##########################################

    context '1 person with on dependent but no counted' do 
      let(:applicant) { create(:applicant, family_situation: 'Célibataire')}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{rel_type: 'droit de visite'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(false)
        expect(hr.reload.category_type).to eq(1)
      end 
    end 

    context 'young couple with 1 children' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 20.years)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}], dependents: [{rel_type: 'E'}] )}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context 'young couple with 1 parent' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 20.years)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}], dependents: [{rel_type: 'P'}] )}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context 'young couple with handicap' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 20.years)}
      let!(:hr) { create(:housing_request,is_active: true, main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}], disability_details: [{lol: 'test'}] )}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.handicap?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context 'young couple with changing date' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 40.years)}
      let!(:hr) { create(:housing_request, is_active: true,  main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}] )}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(false)
        expect(hr.reload.category_type).to eq(2)
        coapplicant.update(birth_date: DateTime.now - 20.years)
        puts "hr.reload.applicant_ids"
        puts hr.reload.applicant_ids
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 
  end 

  describe 'inheriht reference_tax_revenues from coaplicants' do
    let!(:applicant) {create(:applicant)}
    let!(:coapplicant) {create(:applicant)}
    let!(:coapplicant2) {create(:applicant)}
    let!(:hr) {create(:housing_request, main_applicant: applicant, applicant_ids: [coapplicant.id.to_s, coapplicant2.id.to_s], is_active: true)}

    it 'without amount_year_minus_two' do
      expect(hr.reload.reference_tax_revenues).to eq(0)
    end

    it 'only applicant amount_year_minus_two' do
      applicant.update(amount_year_minus_two: 1000)
      expect(hr.reload.reference_tax_revenues).to eq(1000)
    end
    
    it 'all applicants amount_year_minus_two' do
      applicant.update(amount_year_minus_two: 1000)
      coapplicant.update(amount_year_minus_two: 2000)
      expect(hr.reload.reference_tax_revenues).to eq(3000)
      coapplicant2.update(amount_year_minus_two: 500)
      expect(hr.reload.reference_tax_revenues).to eq(3500)
    end

    it 'coapplicant affected after hr creation' do
      coapp = Hermes::Applicant.create(amount_year_minus_two: 800)
      hr.update(applicant_ids: [coapp.id])
      expect(hr.reload.reference_tax_revenues).to eq(800)
    end

  end


  describe 'housing_channel_keys' do 
    let!(:dalo){ create(:housing_channel, key: 'dalo')}
    let!(:lol){ create(:housing_channel, key: 'lol')}
    let!(:applicant){ create(:applicant)}
    let!(:request1) { create(:housing_request, main_applicant: applicant)}
    let!(:request2) { create(:housing_request, main_applicant: applicant, housing_channel_keys: ['dalo'])}

    it 'create with and without housing_channel_keys' do 
      expect(request1.housing_channel_keys).to eq(['dc'])
      expect(request2.housing_channel_keys).to eq(['dalo', 'dc'])
    end 

    it 'when delete housing_channel_keys' do 
      request2.update(housing_channel_keys: [])
      expect(request2.reload.housing_channel_keys).to eq(['dc'])
    end 

    it 'when add housing_channel_keys' do 
      request1.update(housing_channel_keys: ['dalo'])
      expect(request1.reload.housing_channel_keys).to eq(['dalo', 'dc'])
    end 

  end 
end
